@@ -1,6 +1,6 @@
 /*
   This file is part of Challenger
-  (C) 2023, 2024 Taler Systems SA
+  (C) 2023, 2024, 2025 Taler Systems SA
 
   Challenger is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -78,6 +78,22 @@ static struct MHD_Daemon *mhd;
  */
 struct CHALLENGER_DatabasePlugin *CH_db;
 
+/**
+ * (external) base URL of this service.
+ */
+char *CH_base_url;
+
+/**
+ * Mustach template for the letter to send.
+ * WARNING: not 0-terminated! Allocated via mmap(), free with munmap()!
+ */
+void *CH_message_template;
+
+/**
+ * Number of bytes in #CH_message_template
+ */
+size_t CH_message_template_len;
+
 /**
  * How long is an individual validation request valid?
  */
@@ -317,7 +333,7 @@ url_handler (void *cls,
       if (found)
       {
         GNUNET_break_op (0);
-        /* TODO: return which methods are allowed ... */
+        /* FIXME-#9424: return which methods are allowed ... */
         return TALER_MHD_reply_static (hc->connection,
                                        MHD_HTTP_METHOD_NOT_ALLOWED,
                                        "text/plain",
@@ -361,6 +377,17 @@ do_shutdown (void *cls)
     GNUNET_SCHEDULER_cancel (mhd_task);
     mhd_task = NULL;
   }
+  if ( (NULL != CH_message_template) &&
+       (MAP_FAILED != CH_message_template) )
+  {
+    if (0 !=
+        munmap (CH_message_template,
+                CH_message_template_len))
+      GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR,
+                           "munmap");
+    CH_message_template = NULL;
+    CH_message_template_len = 0;
+  }
   if (NULL != CH_ctx)
   {
     GNUNET_CURL_fini (CH_ctx);
@@ -411,14 +438,17 @@ handle_mhd_completion_callback (void *cls,
   if (NULL == hc)
     return;
   GNUNET_assert (hc->connection == connection);
+  GNUNET_break (GNUNET_OK ==
+                CH_db->preflight (CH_db->cls));
   ci = MHD_get_connection_info (connection,
                                 MHD_CONNECTION_INFO_HTTP_STATUS);
   GNUNET_log (GNUNET_ERROR_TYPE_INFO,
               "Finished handling request with status %d (HTTP status %u)\n",
               (int) toe,
-              ci->http_status);
+              (NULL != ci) ? ci->http_status : 0);
   if (NULL != hc->cc)
     hc->cc (hc->ctx);
+  GNUNET_free (hc->full_url);
   GNUNET_free (hc);
   *con_cls = NULL;
 }
@@ -588,6 +618,8 @@ run (void *cls,
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "CHALLENGER",
                                "VALIDATION_DURATION");
+    global_ret = EXIT_NOTCONFIGURED;
+    GNUNET_SCHEDULER_shutdown ();
     return;
   }
   if (GNUNET_OK !=
@@ -599,17 +631,21 @@ run (void *cls,
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "CHALLENGER",
                                "VALIDATION_EXPIRATION");
+    global_ret = EXIT_NOTCONFIGURED;
+    GNUNET_SCHEDULER_shutdown ();
     return;
   }
   if (GNUNET_OK !=
-      GNUNET_CONFIGURATION_get_value_filename (config,
-                                               "CHALLENGER",
-                                               "AUTH_COMMAND",
-                                               &CH_auth_command))
+      GNUNET_CONFIGURATION_get_value_string (config,
+                                             "CHALLENGER",
+                                             "AUTH_COMMAND",
+                                             &CH_auth_command))
   {
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "CHALLENGER",
                                "AUTH_COMMAND");
+    global_ret = EXIT_NOTCONFIGURED;
+    GNUNET_SCHEDULER_shutdown ();
     return;
   }
   if (GNUNET_OK !=
@@ -621,8 +657,101 @@ run (void *cls,
     GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
                                "CHALLENGER",
                                "ADDRESS_TYPE");
+    global_ret = EXIT_NOTCONFIGURED;
+    GNUNET_SCHEDULER_shutdown ();
     return;
   }
+  if (GNUNET_OK !=
+      GNUNET_CONFIGURATION_get_value_string (config,
+                                             "CHALLENGER",
+                                             "BASE_URL",
+                                             &CH_base_url))
+  {
+    GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
+                               "CHALLENGER",
+                               "BASE_URL");
+    global_ret = EXIT_NOTCONFIGURED;
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
+  if ( (! TALER_url_valid_charset (CH_base_url)) ||
+       (! TALER_is_web_url (CH_base_url) ) ||
+       (0 == strlen (CH_base_url)) ||
+       ('/' != CH_base_url[strlen (CH_base_url) - 1]) )
+  {
+    GNUNET_log_config_invalid (GNUNET_ERROR_TYPE_ERROR,
+                               "CHALLENGER",
+                               "BASE_URL",
+                               "invalid URL");
+    global_ret = EXIT_NOTCONFIGURED;
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
+  {
+    char *tmpl_file;
+    int fd;
+    struct stat s;
+
+    if (GNUNET_OK !=
+        GNUNET_CONFIGURATION_get_value_filename (config,
+                                                 "CHALLENGER",
+                                                 "MESSAGE_TEMPLATE_FILE",
+                                                 &tmpl_file))
+    {
+      GNUNET_log_config_missing (GNUNET_ERROR_TYPE_WARNING,
+                                 "CHALLENGER",
+                                 "MESSAGE_TEMPLATE_FILE");
+    }
+    else
+    {
+      fd = open (tmpl_file,
+                 O_RDONLY);
+      GNUNET_free (tmpl_file);
+      if (-1 == fd)
+      {
+        GNUNET_log_config_invalid (GNUNET_ERROR_TYPE_ERROR,
+                                   "CHALLENGER",
+                                   "MESSAGE_TEMPLATE_FILE",
+                                   strerror (errno));
+        global_ret = EXIT_NOTCONFIGURED;
+        GNUNET_SCHEDULER_shutdown ();
+        return;
+      }
+      if (0 != fstat (fd,
+                      &s))
+      {
+        GNUNET_log_config_invalid (GNUNET_ERROR_TYPE_ERROR,
+                                   "CHALLENGER",
+                                   "MESSAGE_TEMPLATE_FILE",
+                                   strerror (errno));
+        GNUNET_free (tmpl_file);
+        GNUNET_break (0 == close (fd));
+        global_ret = EXIT_NOTCONFIGURED;
+        GNUNET_SCHEDULER_shutdown ();
+        return;
+      }
+      CH_message_template_len = (size_t) s.st_size;
+      CH_message_template = mmap (NULL,
+                                  CH_message_template_len,
+                                  PROT_READ,
+                                  MAP_SHARED,
+                                  fd,
+                                  0);
+      GNUNET_break (0 == close (fd));
+      if (MAP_FAILED == CH_message_template)
+      {
+        GNUNET_log_config_invalid (GNUNET_ERROR_TYPE_ERROR,
+                                   "CHALLENGER",
+                                   "MESSAGE_TEMPLATE_FILE",
+                                   strerror (errno));
+        GNUNET_free (tmpl_file);
+        global_ret = EXIT_NOTCONFIGURED;
+        GNUNET_SCHEDULER_shutdown ();
+        return;
+      }
+    }
+  }
+
   {
     char *restrictions;
 
@@ -644,6 +773,8 @@ run (void *cls,
                                    "CHALLENGER",
                                    "ADDRESS_RESTRICTIONS",
                                    err.text);
+        global_ret = EXIT_NOTCONFIGURED;
+        GNUNET_SCHEDULER_shutdown ();
         return;
       }
     }
