@@ -28,6 +28,15 @@
 #include <taler/taler_signatures.h>
 #include "challenger-httpd_common.h"
 
+/**
+ * Maximum number of retries for the database interaction.
+ */
+#define MAX_RETRIES 3
+
+/**
+ * Set to 1 to dump addresses into the log.
+ */
+#define DEBUG 0
 
 /**
  * Context for a /challenge operation.
@@ -95,6 +104,11 @@ struct ChallengeContext
    */
   char *state;
 
+  /**
+   * Buffer used by #TALER_MHD_parse_post_json(). Or NULL.
+   */
+  void *jbuffer;
+
   /**
    * When did we transmit last?
    */
@@ -160,6 +174,11 @@ struct ChallengeContext
    * Did we do the DB interaction?
    */
   bool db_finished;
+
+  /**
+   * Is the upload in JSON?
+   */
+  bool is_json;
 };
 
 
@@ -218,6 +237,7 @@ cleanup_ctx (void *cls)
                   GNUNET_OS_process_wait (bc->child));
     bc->child = NULL;
   }
+  TALER_MHD_parse_post_cleanup_callback (bc->jbuffer);
   json_decref (bc->address);
   GNUNET_free (bc->data);
   GNUNET_free (bc->state);
@@ -267,19 +287,26 @@ child_done_cb (void *cls,
                long unsigned int exit_code)
 {
   struct ChallengeContext *bc = cls;
+  struct GNUNET_AsyncScopeSave old_scope;
 
+  GNUNET_async_scope_enter (&bc->hc->async_scope_id,
+                            &old_scope);
   GNUNET_log (GNUNET_ERROR_TYPE_INFO,
-              "Child done\n");
+              "Child done with exit code %d/%llu\n",
+              (int) type,
+              (unsigned long long) exit_code);
   GNUNET_OS_process_destroy (bc->child);
   bc->child = NULL;
   bc->cwh = NULL;
   bc->pst = type;
   bc->exit_code = exit_code;
+  bc->suspended = GNUNET_NO;
   MHD_resume_connection (bc->hc->connection);
   GNUNET_CONTAINER_DLL_remove (bc_head,
                                bc_tail,
                                bc);
   CH_trigger_daemon ();
+  GNUNET_async_scope_restore (&old_scope);
 }
 
 
@@ -293,7 +320,8 @@ send_tan (struct ChallengeContext *bc)
 {
   struct GNUNET_DISK_PipeHandle *p;
   struct GNUNET_DISK_FileHandle *pipe_stdin;
-  char *msg;
+  void *msg;
+  size_t msg_len;
 
   p = GNUNET_DISK_pipe (GNUNET_DISK_PF_BLOCKING_RW);
   if (NULL == p)
@@ -316,10 +344,12 @@ send_tan (struct ChallengeContext *bc)
 
     address = json_dumps (bc->address,
                           JSON_COMPACT);
+#if DEBUG
     GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                 "Running auth command `%s' on address `%s'\n",
                 CH_auth_command,
                 address);
+#endif
     bc->child = GNUNET_OS_start_process (GNUNET_OS_INHERIT_STD_ERR,
                                          p,
                                          NULL,
@@ -352,13 +382,52 @@ send_tan (struct ChallengeContext *bc)
   GNUNET_assert (NULL != pipe_stdin);
   GNUNET_break (GNUNET_OK ==
                 GNUNET_DISK_pipe_close (p));
+  if (0 != CH_message_template_len)
+  {
+    json_t *root;
+    int mret;
 
-  GNUNET_asprintf (&msg,
-                   "PIN: %u",
-                   (unsigned int) bc->tan);
+    root = GNUNET_JSON_PACK (
+      GNUNET_JSON_pack_object_incref ("address",
+                                      bc->address),
+      GNUNET_JSON_pack_int64 ("pin",
+                              bc->tan));
+    mret = TALER_TEMPLATING_fill2 (CH_message_template,
+                                   CH_message_template_len,
+                                   root,
+                                   &msg,
+                                   &msg_len);
+    json_decref (root);
+    if (0 != mret)
+    {
+      MHD_RESULT mres;
+
+      GNUNET_break (0);
+      mres = reply_error (bc,
+                          "internal-error",
+                          MHD_HTTP_INTERNAL_SERVER_ERROR,
+                          TALER_EC_GENERIC_FAILED_TO_EXPAND_TEMPLATE,
+                          NULL);
+      GNUNET_DISK_file_close (pipe_stdin);
+      bc->status = (MHD_YES == mres)
+          ? GNUNET_NO
+          : GNUNET_SYSERR;
+      return;
+    }
+  }
+  else
+  {
+    char *cmsg;
+
+    GNUNET_asprintf (&cmsg,
+                     "PIN: %u",
+                     (unsigned int) bc->tan);
+    msg_len = strlen (cmsg);
+    msg = cmsg;
+  }
   {
     const char *off = msg;
-    size_t left = strlen (off);
+    size_t left = msg_len;
 
     while (0 != left)
     {
@@ -377,6 +446,7 @@ send_tan (struct ChallengeContext *bc)
                             MHD_HTTP_INTERNAL_SERVER_ERROR,
                             TALER_EC_CHALLENGER_HELPER_EXEC_FAILED,
                             "write");
+        GNUNET_DISK_file_close (pipe_stdin);
         GNUNET_free (msg);
         bc->status = (MHD_YES == mres)
           ? GNUNET_NO
@@ -393,6 +463,7 @@ send_tan (struct ChallengeContext *bc)
                                &child_done_cb,
                                bc);
   MHD_suspend_connection (bc->hc->connection);
+  bc->suspended = GNUNET_YES;
   GNUNET_CONTAINER_DLL_insert (bc_head,
                                bc_tail,
                                bc);
@@ -485,6 +556,9 @@ check_restrictions (const json_t *address)
         "regex"));
     regex_t re;
 
+    if (0 == strcasecmp (key,
+                         "read_only"))
+      continue;
     if (NULL == str)
       return key;
     if (NULL == regex)
@@ -534,7 +608,6 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
     hc->cc = &cleanup_ctx;
     hc->ctx = bc;
     bc->pst = GNUNET_OS_PROCESS_UNKNOWN;
-    bc->address = json_object ();
     bc->tan
       = GNUNET_CRYPTO_random_u32 (GNUNET_CRYPTO_QUALITY_NONCE,
                                   100000000);
@@ -555,6 +628,21 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
                           TALER_EC_GENERIC_PARAMETER_MISSING,
                           hc->path);
     }
+    {
+      const char *ct;
+
+      ct = MHD_lookup_connection_value (hc->connection,
+                                        MHD_HEADER_KIND,
+                                        MHD_HTTP_HEADER_CONTENT_TYPE);
+      bc->is_json = ( (NULL != ct) &&
+                      (0 == strcasecmp (ct,
+                                        "application/json")) );
+      if (! bc->is_json)
+        bc->address = json_object ();
+      GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                  "Processing /challenge upload with %s encoding...\n",
+                  ct);
+    }
     TALER_MHD_check_content_length (hc->connection,
                                     1024);
     GNUNET_log (GNUNET_ERROR_TYPE_INFO,
@@ -570,17 +658,28 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
     return MHD_NO;
   }
   /* Handle case where helper process failed */
-  if ( ( (GNUNET_OS_PROCESS_UNKNOWN != bc->pst) &&
-         (GNUNET_OS_PROCESS_EXITED != bc->pst) ) ||
-       (0 != bc->exit_code) )
+  GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+              "Checking child process status: %d / %d (%s)\n",
+              (int) bc->pst,
+              (int) bc->exit_code,
+              NULL == bc->child ? "dead" : "running");
+  if ( (GNUNET_OS_PROCESS_UNKNOWN != bc->pst) &&
+       (NULL == bc->child) &&
+       ( (GNUNET_OS_PROCESS_EXITED != bc->pst) ||
+         (0 != bc->exit_code) ) )
   {
     char es[32];
 
+    GNUNET_break (0);
     GNUNET_snprintf (es,
                      sizeof (es),
                      "%u/%d",
                      (unsigned int) bc->exit_code,
                      bc->pst);
+    GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
+                "Helper failed with status %d/%d\n",
+                (int) bc->pst,
+                (int) bc->exit_code);
     return reply_error (bc,
                         "internal-error",
                         MHD_HTTP_INTERNAL_SERVER_ERROR,
@@ -588,18 +687,38 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
                         es);
   }
   /* handle upload */
-  if (0 != *upload_data_size)
+  if (bc->is_json)
+  {
+    if (NULL == bc->address)
+    {
+      enum GNUNET_GenericReturnValue res;
+
+      res = TALER_MHD_parse_post_json (hc->connection,
+                                       &bc->jbuffer,
+                                       upload_data,
+                                       upload_data_size,
+                                       &bc->address);
+      if (GNUNET_SYSERR != res)
+        return MHD_YES;
+      GNUNET_break (0);
+      return MHD_NO;
+    }
+    else
+    {
+      GNUNET_break (0 == *upload_data_size);
+    }
+  }
+  else if (0 != *upload_data_size)
   {
     enum MHD_Result res;
 
-    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
-                "Processing /challenge upload...\n");
     res = MHD_post_process (bc->pp,
                             upload_data,
                             *upload_data_size);
     *upload_data_size = 0;
     if (MHD_YES == res)
       return MHD_YES;
+    GNUNET_break (0);
     return MHD_NO;
   }
   if (NULL != bc->last_key)
@@ -612,6 +731,7 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
     bc->data_len = 0;
     GNUNET_free (bc->last_key);
   }
+#if DEBUG
   {
     char *address;
 
@@ -622,6 +742,7 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
                 address);
     free (address);
   }
+#endif
   {
     const char *bad_field;
 
@@ -638,45 +759,109 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
   }
   if (! bc->db_finished)
   {
-    enum GNUNET_DB_QueryStatus qs;
-
-    GNUNET_assert (NULL == bc->client_redirect_uri);
-    qs = CH_db->challenge_set_address_and_pin (
-      CH_db->cls,
-      &bc->nonce,
-      bc->address,
-      CH_validation_duration,
-      &bc->tan,
-      &bc->state,
-      &bc->last_tx_time,
-      &bc->pin_attempts_left,
-      &bc->retransmit,
-      &bc->client_redirect_uri,
-      &bc->address_refused,
-      &bc->solved);
-    switch (qs)
+    for (unsigned int r = 0; r < MAX_RETRIES; r++)
     {
-    case GNUNET_DB_STATUS_HARD_ERROR:
-      GNUNET_break (0);
-      return reply_error (bc,
-                          "internal-error",
-                          MHD_HTTP_INTERNAL_SERVER_ERROR,
-                          TALER_EC_GENERIC_DB_STORE_FAILED,
-                          "set-address-and-pin");
-    case GNUNET_DB_STATUS_SOFT_ERROR:
-      GNUNET_break (0);
-      return MHD_NO;
-    case GNUNET_DB_STATUS_SUCCESS_NO_RESULTS:
-      GNUNET_break_op (0);
-      return reply_error (bc,
-                          "validation-unknown",
-                          MHD_HTTP_NOT_FOUND,
-                          TALER_EC_CHALLENGER_GENERIC_VALIDATION_UNKNOWN,
-                          NULL);
-    case GNUNET_DB_STATUS_SUCCESS_ONE_RESULT:
+      enum GNUNET_DB_QueryStatus qs;
+      json_t *old_address;
+      const json_t *ro;
+
+      GNUNET_assert (NULL == bc->client_redirect_uri);
+      qs = CH_db->address_get (
+        CH_db->cls,
+        &bc->nonce,
+        &old_address);
+      switch (qs)
+      {
+      case GNUNET_DB_STATUS_HARD_ERROR:
+        GNUNET_break (0);
+        return reply_error (bc,
+                            "internal-error",
+                            MHD_HTTP_INTERNAL_SERVER_ERROR,
+                            TALER_EC_GENERIC_DB_FETCH_FAILED,
+                            "validation-get");
+      case GNUNET_DB_STATUS_SOFT_ERROR:
+        if (r < MAX_RETRIES - 1)
+          continue;
+        GNUNET_break (0);
+        return reply_error (bc,
+                            "internal-error",
+                            MHD_HTTP_INTERNAL_SERVER_ERROR,
+                            TALER_EC_GENERIC_DB_FETCH_FAILED,
+                            "validation-get");
+      case GNUNET_DB_STATUS_SUCCESS_NO_RESULTS:
+        GNUNET_break_op (0);
+        return reply_error (bc,
+                            "validation-unknown",
+                            MHD_HTTP_NOT_FOUND,
+                            TALER_EC_CHALLENGER_GENERIC_VALIDATION_UNKNOWN,
+                            NULL);
+      case GNUNET_DB_STATUS_SUCCESS_ONE_RESULT:
+        break;
+      }
+      ro = json_object_get (old_address,
+                            "read_only");
+      if ( (NULL != ro) &&
+           (json_boolean_value (ro)) &&
+           (1 != json_equal (old_address,
+                             bc->address)) )
+      {
+        GNUNET_break_op (0);
+        json_decref (old_address);
+        return reply_error (bc,
+                            "address-read-only",
+                            MHD_HTTP_FORBIDDEN,
+                            TALER_EC_CHALLENGER_CLIENT_FORBIDDEN_READ_ONLY,
+                            NULL);
+      }
+      json_decref (old_address);
+
+      qs = CH_db->challenge_set_address_and_pin (
+        CH_db->cls,
+        &bc->nonce,
+        bc->address,
+        CH_validation_duration,
+        &bc->tan,
+        &bc->state,
+        &bc->last_tx_time,
+        &bc->pin_attempts_left,
+        &bc->retransmit,
+        &bc->client_redirect_uri,
+        &bc->address_refused,
+        &bc->solved);
+      switch (qs)
+      {
+      case GNUNET_DB_STATUS_HARD_ERROR:
+        GNUNET_break (0);
+        return reply_error (bc,
+                            "internal-error",
+                            MHD_HTTP_INTERNAL_SERVER_ERROR,
+                            TALER_EC_GENERIC_DB_STORE_FAILED,
+                            "set-address-and-pin");
+      case GNUNET_DB_STATUS_SOFT_ERROR:
+        if (r < MAX_RETRIES - 1)
+          continue;
+        GNUNET_break (0);
+        return reply_error (bc,
+                            "internal-error",
+                            MHD_HTTP_INTERNAL_SERVER_ERROR,
+                            TALER_EC_GENERIC_DB_STORE_FAILED,
+                            "set-address-and-pin");
+      case GNUNET_DB_STATUS_SUCCESS_NO_RESULTS:
+        GNUNET_break_op (0);
+        return reply_error (bc,
+                            "validation-unknown",
+                            MHD_HTTP_NOT_FOUND,
+                            TALER_EC_CHALLENGER_GENERIC_VALIDATION_UNKNOWN,
+                            NULL);
+      case GNUNET_DB_STATUS_SUCCESS_ONE_RESULT:
+        break;
+      }
       break;
     }
     bc->db_finished = true;
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Database interaction done (challenge %s)\n",
+                bc->solved ? "solved" : "unsolved");
     if (bc->solved)
     {
       struct MHD_Response *response;
@@ -726,7 +911,11 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
       /* (Re)transmit PIN/TAN */
       send_tan (bc);
       if (GNUNET_YES == bc->suspended)
+      {
+        GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                    "Suspending request after PIN transmission\n");
         return MHD_YES;
+      }
       /* Did we already try to generate a response? */
       if (GNUNET_OK != bc->status)
         return (GNUNET_NO == bc->status)
@@ -741,11 +930,12 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
     unsigned int http_status;
     MHD_RESULT res;
 
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Returning success from challenge with %u attempts left\n",
+                (unsigned int) bc->pin_attempts_left);
     args = GNUNET_JSON_PACK (
       GNUNET_JSON_pack_uint64 ("attempts_left",
                                bc->pin_attempts_left),
-      GNUNET_JSON_pack_string ("nonce",
-                               hc->path),
       GNUNET_JSON_pack_string ("type",
                                "created"),
       GNUNET_JSON_pack_object_incref ("address",
@@ -759,10 +949,7 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
                                       CH_validation_duration)))
       );
     http_status = MHD_HTTP_OK;
-    GNUNET_break (0 ==
-                  json_object_del (args,
-                                   "nonce"));
-    resp = TALER_MHD_make_json (args);
+    resp = TALER_MHD_make_json_steal (args);
     GNUNET_break (MHD_YES ==
                   MHD_add_response_header (resp,
                                            MHD_HTTP_HEADER_CACHE_CONTROL,
