@@ -27,6 +27,10 @@
 #include <taler/taler_templating_lib.h>
 #include <taler/taler_signatures.h>
 
+/**
+ * Maximum number of retries for the database interaction.
+ */
+#define MAX_RETRIES 3
 
 /**
  * Context for a /solve operation.
@@ -265,37 +269,49 @@ CH_handler_solve (struct CH_HandlerContext *hc,
                           "pin");
     }
 
-    qs = CH_db->validate_solve_pin (CH_db->cls,
-                                    &bc->nonce,
-                                    pin,
-                                    &solved,
-                                    &exhausted,
-                                    &no_challenge,
-                                    &bc->state,
-                                    &bc->addr_left,
-                                    &bc->auth_attempts_left,
-                                    &bc->pin_transmissions_left,
-                                    &bc->client_redirect_uri);
-    switch (qs)
+    for (unsigned int r = 0; r<MAX_RETRIES; r++)
     {
-    case GNUNET_DB_STATUS_HARD_ERROR:
-      return reply_error (
-        bc,
-        "internal-error",
-        MHD_HTTP_INTERNAL_SERVER_ERROR,
-        TALER_EC_GENERIC_DB_FETCH_FAILED,
-        "validate_solve_pin");
-    case GNUNET_DB_STATUS_SOFT_ERROR:
-      GNUNET_break (0);
-      return MHD_NO;
-    case GNUNET_DB_STATUS_SUCCESS_NO_RESULTS:
-      return reply_error (
-        bc,
-        "validation-unknown",
-        MHD_HTTP_NOT_FOUND,
-        TALER_EC_CHALLENGER_GENERIC_VALIDATION_UNKNOWN,
-        NULL);
-    case GNUNET_DB_STATUS_SUCCESS_ONE_RESULT:
+      qs = CH_db->validate_solve_pin (CH_db->cls,
+                                      &bc->nonce,
+                                      pin,
+                                      &solved,
+                                      &exhausted,
+                                      &no_challenge,
+                                      &bc->state,
+                                      &bc->addr_left,
+                                      &bc->auth_attempts_left,
+                                      &bc->pin_transmissions_left,
+                                      &bc->client_redirect_uri);
+      switch (qs)
+      {
+      case GNUNET_DB_STATUS_HARD_ERROR:
+        GNUNET_break (0);
+        return reply_error (
+          bc,
+          "internal-error",
+          MHD_HTTP_INTERNAL_SERVER_ERROR,
+          TALER_EC_GENERIC_DB_FETCH_FAILED,
+          "validate_solve_pin");
+      case GNUNET_DB_STATUS_SOFT_ERROR:
+        if (r < MAX_RETRIES - 1)
+          continue;
+        GNUNET_break (0);
+        return reply_error (
+          bc,
+          "internal-error",
+          MHD_HTTP_INTERNAL_SERVER_ERROR,
+          TALER_EC_GENERIC_DB_FETCH_FAILED,
+          "validate_solve_pin");
+      case GNUNET_DB_STATUS_SUCCESS_NO_RESULTS:
+        return reply_error (
+          bc,
+          "validation-unknown",
+          MHD_HTTP_NOT_FOUND,
+          TALER_EC_CHALLENGER_GENERIC_VALIDATION_UNKNOWN,
+          NULL);
+      case GNUNET_DB_STATUS_SUCCESS_ONE_RESULT:
+        break;
+      }
       break;
     }
     if (! solved)
@@ -319,11 +335,11 @@ CH_handler_solve (struct CH_HandlerContext *hc,
       GNUNET_log (GNUNET_ERROR_TYPE_INFO,
                   "Invalid PIN supplied\n");
       details = GNUNET_JSON_PACK (
+        GNUNET_JSON_pack_string ("type",
+                                 "pending"),
         TALER_JSON_pack_ec (TALER_EC_CHALLENGER_INVALID_PIN),
         GNUNET_JSON_pack_uint64 ("addresses_left",
                                  bc->addr_left),
-        GNUNET_JSON_pack_string ("type",
-                                 "pending"),
         GNUNET_JSON_pack_uint64 ("pin_transmissions_left",
                                  bc->pin_transmissions_left),
         GNUNET_JSON_pack_uint64 ("auth_attempts_left",
