@@ -1,9 +1,8 @@
 #!/bin/sh
 # This file is in the public domain.
 set -eu
-# . telesign-secrets
 # Set AUTH_TOKEN=...
-NUMBER=$(echo "$1" | jq -r .phone)
+NUMBER=$(echo "$1" | jq -r .CONTACT_PHONE)
 MESSAGE=$(cat -)
 TMPFILE=$(mktemp /tmp/sms-loggingXXXXXX)
 STATUS=$(curl --request "POST" \
