@@ -49,20 +49,10 @@ static const struct GNUNET_OS_ProjectData challenger_pd = {
 
 
 const struct GNUNET_OS_ProjectData *
-CHALLENGER_project_data_default (void)
+CHALLENGER_project_data (void)
 {
   return &challenger_pd;
 }
 
 
-/**
- * Initialize libchallengerutil.
- */
-void __attribute__ ((constructor))
-CHALLENGER_OS_init ()
-{
-  GNUNET_OS_init (&challenger_pd);
-}
-
-
 /* end of os_installation.c */
